/**
 * vue-stator v1.1.1
 * (c) 2019
 * - Jonas Galvez (@galvez)
 * - Pim (@pimlie)
 * - All the amazing contributors
 * @license MIT
 */

import Vue from 'vue';

function _typeof(obj) {
  if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") {
    _typeof = function (obj) {
      return typeof obj;
    };
  } else {
    _typeof = function (obj) {
      return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
    };
  }

  return _typeof(obj);
}

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _nonIterableRest();
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArrayLimit(arr, i) {
  if (!(Symbol.iterator in Object(arr) || Object.prototype.toString.call(arr) === "[object Arguments]")) {
    return;
  }

  var _arr = [];
  var _n = true;
  var _d = false;
  var _e = undefined;

  try {
    for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance");
}

var namespaceSeparator = '/';
function validateNamespace(namespace) {
  if (typeof namespace === 'string') {
    namespace = namespace.split(namespaceSeparator);
  }

  if (!Array.isArray(namespace)) {
    // TODO: warn?
    return [];
  }

  return namespace.filter(Boolean);
}
function getPropertyByNamespace(object, namespace) {
  namespace = validateNamespace(namespace);
  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = namespace[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var name = _step.value;

      if (name in object) {
        object = object[name];
        continue;
      } // TODO: warn on else?

    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return != null) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  return object;
}
function setPropertyByNamespace(object, namespace, value) {
  namespace = validateNamespace(namespace);
  var key = namespace.pop();
  var _iteratorNormalCompletion2 = true;
  var _didIteratorError2 = false;
  var _iteratorError2 = undefined;

  try {
    for (var _iterator2 = namespace[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
      var name = _step2.value;

      if (name in object) {
        object = object[name];
        continue;
      } // TODO: warn on else?

    }
  } catch (err) {
    _didIteratorError2 = true;
    _iteratorError2 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion2 && _iterator2.return != null) {
        _iterator2.return();
      }
    } finally {
      if (_didIteratorError2) {
        throw _iteratorError2;
      }
    }
  }

  object[key] = value;
  return object[key];
}
function getModuleByNamespace(store, namespace) {
  namespace = validateNamespace(namespace);
  var object = {
    state: store.$state,
    getters: store.$getters,
    actions: store.$actions
  };
  var _iteratorNormalCompletion3 = true;
  var _didIteratorError3 = false;
  var _iteratorError3 = undefined;

  try {
    for (var _iterator3 = namespace[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
      var name = _step3.value;

      if (!object.state[name]) {
        object.state[name] = {};
      }

      if (!object.getters[name]) {
        object.getters[name] = {};
      }

      if (!object.actions[name]) {
        object.actions[name] = {};
      }

      object = {
        state: object.state[name],
        getters: object.getters[name],
        actions: object.actions[name]
      };
    }
  } catch (err) {
    _didIteratorError3 = true;
    _iteratorError3 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion3 && _iterator3.return != null) {
        _iterator3.return();
      }
    } finally {
      if (_didIteratorError3) {
        throw _iteratorError3;
      }
    }
  }

  return object;
}

function ensureArray(variable) {
  if (Array.isArray(variable)) {
    return variable;
  }

  return [variable];
}
function callIfFunction(variable) {
  if (typeof variable === 'function') {
    for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      args[_key - 1] = arguments[_key];
    }

    return variable.apply(void 0, args);
  }

  return variable;
}

function $set() {
  Vue.set.apply(Vue, arguments);
}
function $delete() {
  Vue.delete.apply(Vue, arguments);
}
function map(properties, assignMapping) {
  var mapped = {};

  if (_typeof(properties) !== 'object') {
    properties = ensureArray(properties);
  }

  if (Array.isArray(properties)) {
    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;
    var _iteratorError = undefined;

    try {
      for (var _iterator = properties[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
        var prop = _step.value;
        mapped[prop] = assignMapping(prop);
      }
    } catch (err) {
      _didIteratorError = true;
      _iteratorError = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion && _iterator.return != null) {
          _iterator.return();
        }
      } finally {
        if (_didIteratorError) {
          throw _iteratorError;
        }
      }
    }

    return mapped;
  }

  for (var propName in properties) {
    mapped[propName] = assignMapping(properties[propName]);
  }

  return mapped;
}
function mapState(namespacedProperties) {
  return map(namespacedProperties, function (prop) {
    return {
      get: function get() {
        return getPropertyByNamespace(this.$state, prop);
      },
      set: function set(value) {
        return setPropertyByNamespace(this.$state, prop, value);
      }
    };
  });
}
function mapGetters(namespacedProperties) {
  return map(namespacedProperties, function (prop) {
    return function () {
      return getPropertyByNamespace(this.$getters, prop);
    };
  });
}
function mapActions(namespacedProperties) {
  return map(namespacedProperties, function (prop) {
    return function () {
      var action = getPropertyByNamespace(this.$actions, prop);

      if (action) {
        return action.apply(void 0, arguments);
      }
    };
  });
}

var logScope = "[vue-stator]";
var logger = console;

var watcher;
function setWatcher(vm) {
  watcher = vm;
}
function subscribe(_ref, namespace, callback, vm) {
  var $state = _ref.$state;

  if (!watcher) {
    setWatcher(vm || new Vue());
  }

  return watcher.$watch(function () {
    return getPropertyByNamespace($state, namespace);
  }, callback, {
    deep: true
  });
}

function validateStorage(storage) {
  if (!storage) {
    return false;
  }

  if (storage instanceof Storage) {
    return true;
  }

  if (typeof storage.getItem !== 'function') {
    return false;
  }

  if (typeof storage.setItem === 'function') {
    return true;
  }

  return false;
} // Internal utilities

function dumpState(storage, namespace, data) {
  /* istanbul ignore next */
  if (!storage) {
    return;
  }

  storage.setItem("vue-stator:".concat(namespace), JSON.stringify(data));
}
function loadState(storage, namespace, onSuccess, onError) {
  /* istanbul ignore next */
  if (!storage) {
    return;
  }

  try {
    var data = JSON.parse(storage.getItem("vue-stator:".concat(namespace)));
    onSuccess(data);
  } catch (err) {
    logger.error("".concat(logScope, " Error in loadState(").concat(Array.prototype.slice.call(arguments).join(', '), ")"), err);

    if (onError) {
      onError(err);
    }
  }
}
function registerStorage(store, storage) {
  if (!Array.isArray(storage)) {
    storage = [storage];
  }

  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    var _loop = function _loop() {
      var shelf = _step.value;
      var provider = void 0;
      var namespaces = void 0;

      if (Array.isArray(shelf)) {
        var _shelf = _slicedToArray(shelf, 2);

        provider = _shelf[0];
        namespaces = _shelf[1];
      } else {
        provider = shelf.storage;
        namespaces = shelf.namespaces;
      }

      provider = callIfFunction(provider);

      if (!validateStorage(provider)) {
        // TODO: warn?
        return "continue";
      } // TODO: multiple depth support


      var _iteratorNormalCompletion2 = true;
      var _didIteratorError2 = false;
      var _iteratorError2 = undefined;

      try {
        var _loop2 = function _loop2() {
          var namespace = _step2.value;
          loadState(provider, namespace, function (data) {
            return $set(store.$state, namespace, data);
          });
          subscribe(store, namespace, function (snapshot) {
            return dumpState(provider, namespace, snapshot);
          });
        };

        for (var _iterator2 = namespaces[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
          _loop2();
        }
      } catch (err) {
        _didIteratorError2 = true;
        _iteratorError2 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion2 && _iterator2.return != null) {
            _iterator2.return();
          }
        } finally {
          if (_didIteratorError2) {
            throw _iteratorError2;
          }
        }
      }
    };

    for (var _iterator = storage[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var _ret = _loop();

      if (_ret === "continue") continue;
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return != null) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }
}

function createStore(_ref, vm) {
  var state = _ref.state,
      context = _ref.context,
      hydrate = _ref.hydrate,
      actions = _ref.actions,
      getters = _ref.getters,
      modules = _ref.modules,
      storage = _ref.storage;

  if (vm) {
    setWatcher(vm);
  }

  var initialState = state ? callIfFunction(state) : {};

  if (hydrate) {
    initialState = hydrate(initialState);
  }

  var store = {
    $state: Vue.observable(initialState),
    $getters: {},
    $actions: {},
    subscribe: function subscribe$1() {
      for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
        args[_key] = arguments[_key];
      }

      return subscribe.apply(void 0, [store].concat(args));
    },
    registerModule: function registerModule(module, namespace) {
      var moduleName = module.name;

      if (!moduleName) {
        throw new Error('module name is required when registering an object');
      }

      delete module.namespaced;

      _registerModule(store, namespace, module, moduleName);
    },
    unregisterModule: function unregisterModule(moduleName) {
      _unregisterModule(store, moduleName);
    }
  };
  context = context || {};
  context.$stator = store;
  context.$state = store.$state;
  context.$getters = store.$getters;
  context.$actions = store.$actions;
  register(context, {
    state: store.$state,
    getters: store.$getters,
    actions: store.$actions
  }, {
    getters: getters,
    actions: actions,
    modules: modules
  }, true);

  if (storage) {
    registerStorage(store, storage);
  }

  return store;
} // Registration helpers

function register(context, parent, _ref2, isRoot) {
  var getters = _ref2.getters,
      actions = _ref2.actions,
      modules = _ref2.modules;
  var store = context.$stator || context;
  registerGetters(store, parent, getters);
  registerActions(context, parent, actions);

  if (!modules) {
    return;
  }

  for (var moduleName in modules) {
    // this is to support a flattened module object
    // TODO: should we order moduleNames by length? Cause now the user needs
    // to register the module 'user' before 'user/profile'
    if (isRoot && moduleName.includes(namespaceSeparator)) {
      _registerModule(context, moduleName, modules[moduleName]);

      continue;
    }

    _registerModule(context, parent, modules[moduleName], moduleName);
  }
}

function _registerModule(context, parent, module, moduleName) {
  // for supporting runtime module registration
  if (typeof parent === 'string') {
    parent = parent.split(namespaceSeparator);
  }

  if (Array.isArray(parent)) {
    var store = context.$stator || context;
    moduleName = moduleName || parent.pop();
    parent = getModuleByNamespace(store, parent);
  } // namespaced is the default, only dont namespace
  // when its explicitly set to falsy


  var moduleParent;

  if (!('namespaced' in module) || module.namespaced) {
    var moduleState = registerState(parent, moduleName, module); // if parent has getter/action with same
    // name as this module, the module has preference
    // TODO: should we warn about this?

    parent.getters[moduleName] = {};
    parent.actions[moduleName] = {};
    moduleParent = {
      state: moduleState,
      getters: parent.getters[moduleName],
      actions: parent.actions[moduleName]
    };
  }

  register(context, moduleParent || parent, module);
}

function _unregisterModule(context, parent) {
  if (typeof parent === 'string') {
    parent = parent.split(namespaceSeparator);
  }

  if (!Array.isArray(parent)) {
    // TODO: throw error?
    return;
  }

  var store = context.$stator || context;
  var moduleName = parent.pop();
  parent = getModuleByNamespace(store, parent);

  for (var _i = 0, _arr = ['state', 'getters', 'actions']; _i < _arr.length; _i++) {
    var key = _arr[_i];

    if (parent[key][moduleName]) {
      delete parent[key][moduleName];
    }
  }
}
function registerState(_ref3, namespace, _ref4) {
  var parent = _ref3.state;
  var state = _ref4.state;

  if (!state) {
    // this allows to use an unified state file
    return parent[namespace] || parent;
  } // TODO: check if not fn on ssr and print warning?
  // TODO: what about hook so users can manually clone the state here?


  Vue.set(parent, namespace, callIfFunction(state));
  return parent[namespace];
}
function registerGetters(store, _ref5, getters) {
  var state = _ref5.state,
      parent = _ref5.getters;

  if (!getters) {
    return;
  }

  var _loop = function _loop(key) {
    if (typeof getters[key] !== 'function') {
      // TODO: huh, why are we here?
      return "continue";
    }

    Object.defineProperty(parent, key, {
      get: function get() {
        return getters[key](state, store.$getters, store.$state);
      }
    });
  };

  for (var key in getters) {
    var _ret = _loop(key);

    if (_ret === "continue") continue;
  }
}
function registerActions(context, _ref6, actions) {
  var state = _ref6.state,
      parent = _ref6.actions;

  if (!actions) {
    return;
  }

  var _loop2 = function _loop2(key) {
    if (typeof actions[key] !== 'function') {
      // TODO: huh, why are we here?
      return "continue";
    } // TODO: why no arrow fn here? do we need this context?


    parent[key] = function () {
      for (var _len2 = arguments.length, args = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
        args[_key2] = arguments[_key2];
      }

      return actions[key].apply(actions, [context, state].concat(args));
    };
  };

  for (var key in actions) {
    var _ret2 = _loop2(key);

    if (_ret2 === "continue") continue;
  }
}

var injectedPropertyName = 'stator';
function injectStator(Vue, options) {
  lazyInject(Vue, injectedPropertyName, function (vm) {
    return createStore(options, vm.$root);
  });
  injectProperty(Vue, injectedPropertyName, '$state');
  injectProperty(Vue, injectedPropertyName, '$getters');
  injectProperty(Vue, injectedPropertyName, '$actions');
}
function lazyInject(Vue, key, setter) {
  Vue.use(function () {
    var propertyName = "$".concat(key);

    if (Vue.prototype.hasOwnProperty(propertyName)) {
      return;
    }

    Object.defineProperty(Vue.prototype, propertyName, {
      configurable: true,
      enumerable: false,
      get: function get() {
        if (!(key in this.$root.$options)) {
          this.$root.$options[key] = setter(this);
        }

        inject(Vue, propertyName, key);
        return this.$root.$options[key];
      }
    });
  });
}
function inject(Vue, propertyName, key) {
  Object.defineProperty(Vue.prototype, propertyName, {
    enumerable: false,
    get: function get() {
      return this.$root.$options[key];
    }
  });
}
function injectProperty(Vue, propertyName, key) {
  if (Vue.prototype.hasOwnProperty(key)) {
    return;
  }

  Object.defineProperty(Vue.prototype, key, {
    enumerable: true,
    get: function get() {
      return this["$".concat(propertyName)][key];
    }
  });
}
var plugin = {
  install: function install(Vue, options) {
    // Check if plugin not already installed
    var installKey = "__stator_installed";

    if (Vue[installKey]) {
      return;
    }

    Vue[installKey] = true;
    injectStator(Vue, options);
  }
};

export { $delete, $set, createStore, map, mapActions, mapGetters, mapState, plugin, register, registerActions, registerGetters, _registerModule as registerModule, registerState, setWatcher, subscribe, _unregisterModule as unregisterModule };
